;; @(#) glide-debug.el -- Extended debugger support for Glide

;; Copyright (C) 1999 Ada Core Technologies

;; Author: Emmanuel Briot <briot@gnat.com>
;; Maintainer: Emmanuel Briot <briot@gnat.com>
;; Ada Core Technologies's version:   $Revision: 1.23 $
;; Keywords: languages ada glide

;; This file is not part of GNU Emacs.

;; This program is free software; you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation; either version 2, or (at your option)
;; any later version.

;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.

;; You should have received a copy of the GNU General Public License
;; along with GNU Emacs; see the file COPYING.  If not, write to
;; the Free Software Foundation, 675 Mass Ave, Cambridge, MA 02139, USA.

;;; Commentary:
;;; This file enhanced the standard debugger mode from Emacs and XEmacs.
;;; It adds a number of new things:
;;;   - complete menus (up to Emacs 20.5, gud.el does not display most of the
;;;     items
;;;   - DDD support (when a variable is displayed, a new item is created in
;;;     ddd's graphical window)
;;;   - enhanced breakpoints: whenever a breakpoint is set or removed, the
;;;     matching line in the source files is highlighted while the breakpoint
;;;     is alive
;;;   - contextual menus: while the debugger is active, the contextual menu
;;;     in the ada-mode gets some new debugger-related items
;;;   - distinguished between the print and the display function (the print
;;;     output is always sent to the debugger buffer, whereas the behavior of
;;;     display depends on the debugger used (gdb, ddd, ...)

;;; Known bugs:
;;;   - temporary breakpoints are not supported: the line is correctly
;;;     hightlighted, but never changed back to normal state.
;;;   - The contextual menu part should be as independant as possible from
;;;     the ada-mode
;;;   - This file can handled only one debugger session at a time

(require 'gud)


(defvar glide-debug-breakpoints '()
  "List of breakpoints defined in the current gdb section.")

(defvar glide-debug-default-filter 'gud-gdb-marker-filter
  "Default filter that was set by the debugger session.")

(defvar glide-debug-buffer nil
  "Buffer that contains the debugger session.")

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  The breakpoint structure
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;; This structure is used internally to represent a breakpoint
;; It contains the following information:
;;   number:  Internal breakpoint number in gdb
;;   file:    File in which the breakpoint is set
;;   line:    Line on which the breakpoint is set
;;   overlay: Overlay used to highlight the line in Emacs

(defmacro glide-debug-make-breakpoint () (make-vector 4 nil))
(defmacro glide-debug-number-of (br) (list 'aref br 0))
(defmacro glide-debug-file-of   (br) (list 'aref br 1))
(defmacro glide-debug-line-of   (br) (list 'aref br 2))
(defmacro glide-debug-overlay-of (br) (list 'aref br 3))

(defmacro glide-debug-set-number (br number) (list 'aset br 0 number))
(defmacro glide-debug-set-file   (br file)   (list 'aset br 1 file))
(defmacro glide-debug-set-line   (br line)   (list 'aset br 2 line))
(defmacro glide-debug-set-overlay (br over)  (list 'aset br 3 over))


;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  DDD Support
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun glide-debug-set-ddd-support ()
  "Add support for the Data Display Debugger to the gdb-mode"
  
  (defun gud-display (arg) (interactive "p")
    (gud-call "graph display %e" arg))
  )

(defun glide-debug-set-gvd-support ()
  "Add support for the Other Display Debugger to the gdb-mode"

  (glide-debug-set-ddd-support)
  (defun gud-print (arg) (interactive "p")
    (gud-call "graph print %e" arg))
  (defun gud-view-source () (interactive "p")
    (gud-call "view source"))
  (defun gud-view-asm () (interactive "p")
    (gud-call "view source_asm"))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  Enhanced breakpoint support
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun glide-debug-remove-breakpoint (breakpoint)
  "The breakpoint BREAKPOINT was removed. This function updates the
highlight status of the corresponding line."

  (if breakpoint
      (let ((overlay (glide-debug-overlay-of breakpoint)))
	(if overlay
	    (delete-overlay overlay)))))

(defun glide-debug-set-breakpoint (line)
  "Set a breakpoint on LINE in the current buffer.
Returns the overlay that was created."
  (save-excursion
    (goto-line line)
    (let ((overlay
	   (make-overlay (point)
			 (save-excursion (end-of-line) (1+ (point))))))
      (overlay-put overlay 'face 'highlight)
      overlay)))


(defun glide-debug-find-file ()
  "Hook called whenever a new file is opened and the debugger mode is active.
It checks whether some lines have some breakpoints and should be highlighted"

  (let ((name (file-name-nondirectory (buffer-file-name)))
	(br   (copy-sequence glide-debug-breakpoints)))

    ;;  If the buffer is in the list of breakpoints
    (while br
      (if (string= (glide-debug-file-of (car br)) name)
	  (glide-debug-set-overlay
	   (car br)
	   (glide-debug-set-breakpoint (glide-debug-line-of (car br)))))
      (set 'br (cdr br)))
  ))


(defun glide-debug-status-changed (proc msg)
  "Called whenever the debug process changes status. This is used to
update the content of the contextual menu for the ada-mode"
  
  (let ((status (process-status proc)))
    (cond ((memq status '(signal stop exit closed))
	   (set 'glide-debug-buffer nil))
	  (t
	   (set 'glide-debug-buffer (process-buffer proc)))))

  ;; Suppress the highlighting for all the existing breakpoints
  (unless glide-debug-buffer
    (progn
      (save-excursion
	(while glide-debug-breakpoints
	  (glide-debug-remove-breakpoint (car glide-debug-breakpoints))
	  (set 'glide-debug-breakpoints (cdr glide-debug-breakpoints))))
      (remove-hook 'find-file-hooks 'glide-debug-find-file)
      ))
  
  (gud-sentinel proc msg))


(defun glide-debug-marker-filter (string)
  "Works like `gud-gdb-marker-filter'. It filters the output of gdb to
detect when a breakpoint was set and when a breakpoint was deleted
Whenever a new breakpoint was created, hightlight the matching line in the
source buffer"

  (let (output)

    (let ((gud-marker-filter glide-debug-default-filter))
      (set 'output (gud-marker-filter string)))
    
    (if (string-match (concat "Breakpoint \\([0-9]+\\) at \\([0-9xa-f]+\\): file"
			      " \\([^,]+\\), line \\([0-9]+\\)")
		      output)
	(let* ((file (match-string 3 output))
	       (line (string-to-number (match-string 4 output)))
	       (number (match-string 1 output))
	       overlay breakpoint
	       (buf    (gud-find-file file)))
 	  (if buf
 	      (save-excursion
 		;; save the position in a list, so that find-file can rehighlight the
 		;; line if the buffer is reloaded
 		(set-buffer (gud-find-file file))
		
 		;; Now highlight the line (add 1 so that the whole line is
 		;; highlighted)
 		(set 'breakpoint (glide-debug-make-breakpoint))
 		(glide-debug-set-overlay breakpoint
 					 (glide-debug-set-breakpoint line))
 		(glide-debug-set-number  breakpoint number)
 		(glide-debug-set-file    breakpoint file)
 		(glide-debug-set-line    breakpoint line)
 		(add-to-list 'glide-debug-breakpoints breakpoint)
 		))
	  ))
    
    ;;  If a breakpoint was deleted
    
    (if (string-match "Deleted breakpoint \\([0-9]+\\)" output)
	(let ((br (copy-sequence glide-debug-breakpoints))
	      (num (match-string 1 output)))
	  (while br
	    (if (string= (glide-debug-number-of (car br)) num)
		(glide-debug-remove-breakpoint (car br)))
	    (set 'br (cdr br)))))
    output))

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

;;  Enhanced menus
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;; For all versions of Emacs, up to 20.5, the menu is incorrectly
;;; set, and most items are not visible (nor in the source buffer).
;;; This is an attempt at correcting things

(defun glide-debug-menus ()
  "Modifies the menu handling functions in gud.el. It basically
redefines a set of functions"

  ;;  the modification of the function is done there, so that the loading
  ;;  of the debugger mode does not interfer with these definitions
  (defun gud-make-debug-menu ()
    (if (current-local-map)
	(define-key ada-mode-map [menu-bar debug]
	  (cons "Debugger" gud-menu-map))))
  )

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  Contextual menus
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun glide-debug-contextual-menu (menu)
  "Add some new items to a menu (intended to be contextual, but this is not
mandatory).
Returns the enhanced menu"

  ;;  The items here a conditionnalized by two variables, glide-debug-buffer
  ;;  and ada-contextual-menu-on-identifier

  (easy-menu-add-item menu nil
		      '["--" nil :included glide-debug-buffer])
  (easy-menu-add-item menu nil
		      '["Debugger Output" (display-buffer glide-debug-buffer t)
			:included glide-debug-buffer])
  (easy-menu-add-item menu nil
		      '["Print Expression" (gud-print 0)
			:included  (and glide-debug-buffer
					ada-contextual-menu-on-identifier)
			:keys "\\[gud-print]"])
  (easy-menu-add-item menu nil
		      '["Display Expression" (gud-display 0)
			:included  (and glide-debug-buffer
					ada-contextual-menu-on-identifier)
			:keys "\\[gud-display]"])
  (easy-menu-add-item menu nil
		      '["What is Expression" (gud-ptype 0)
			:included  (and glide-debug-buffer
					ada-contextual-menu-on-identifier)
			:keys "\\[gud-ptype]"])
  (easy-menu-add-item menu nil
		      '["Breakpoint" (gud-break 0)
			:included  glide-debug-buffer
			:keys "\\[gud-break]"])
  (easy-menu-add-item menu nil
		      '["Temporary Breakpoint" (gud-tbreak 0)
			:included  glide-debug-buffer
			:keys "\\[gud-tbreak]"])
  (easy-menu-add-item menu nil
		      '["Remove Breakpoint" (gud-remove 0)
			:included  glide-debug-buffer
			:keys "\\[gud-remove]"])
  (easy-menu-add-item menu nil
		      '["Up Stack" (gud-up 1)
			:included  glide-debug-buffer
			:keys "\\[gud-up]"])
  (easy-menu-add-item menu nil
		      '["Down Stack"  (gud-down 1)
			:included  glide-debug-buffer
			:keys "\\[gud-down]"])
  (easy-menu-add-item menu nil
		      '["Next Line"  (gud-next 1)
			:included  glide-debug-buffer
			:keys "\\[gud-next]"])
  (easy-menu-add-item menu nil
		      '["Step Line"  (gud-step 1)
			:included  glide-debug-buffer
			:keys "\\[gud-step]"])
  (easy-menu-add-item menu nil
		      '["Continue"  (gud-cont 0)
			:included  glide-debug-buffer
			:keys "\\[gud-cont]"])
  (easy-menu-add-item menu nil
		      '["Run" (gud-run 0)
			:included  glide-debug-buffer
			:keys "\\[gud-run]"])
  (easy-menu-add-item menu nil
		      '["View Asm" (gud-view-asm)
			:included  glide-debug-buffer
			:keys "\\[gud-view-asm]"])
  (easy-menu-add-item menu nil
		      '["View Source" (gud-view-source)
			:included  glide-debug-buffer
			:keys "\\[gud-view-source]"])
  menu)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  Overrides some of the standard C features
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defadvice gud-call (before glide-gud-call-advice)
  "Substitute the %e in debugger commands for Ada files.
It then calls `gud-call' to replace the other special substrings."
  (if (string= mode-name "Ada")
      (let (subst)

	;;  If the region is active, use it
	(if (or (and ada-xemacs (region-active-p))
		(and (not ada-xemacs)
		     transient-mark-mode
		     mark-active))
	    (save-excursion
	      (goto-char (car ada-contextual-menu-last-point))
	      (set 'subst (buffer-substring-no-properties (region-beginning)
							  (region-end))))

	  ;;  Else get the identifier under the cursor
	  (set 'subst (buffer-substring-no-properties
		       (save-excursion (skip-chars-backward "a-zA-Z0-9_")
				       (point))
		       (save-excursion (skip-chars-forward "a-zA-Z0-9_")
				       (point))))
	  )
	
	(while (and fmt (string-match "\\(%e\\)" fmt))
	  (set 'fmt (replace-match
		     subst
		     t t fmt))
	  ))))

(ad-activate 'gud-call t)

;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;
;;  Initializes the Glide-Debug mode
;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;;

(defun glide-debug-initialize ()
  "Initializes and activates the extended debugger support found
This is done for the current debugger session.
When calling this function, the debugger buffer should be the current one"

  (let (process)
    ;;  Get the gdb process that is executed
    (set 'process (get-buffer-process (current-buffer)))
    (if (listp process)
	(set 'process (car process)))

    (set-process-sentinel process 'glide-debug-status-changed)

    ;;  Modifies the standard gud menu, so that all functions appear in there
    (if (boundp 'gud-menu-map)
	(progn
	  (define-key gud-menu-map [tbreak]
	    '("Temporary Breakpoint".gud-tbreak))
	  (define-key gud-menu-map [finish]  '("Finish Function" . gud-finish))
	  (define-key gud-menu-map [up]      '("Up Stack" . gud-up))
	  (define-key gud-menu-map [down]    '("Down Stack" . gud-down))
	  (define-key gud-menu-map [ptype]   '("What is" . gud-ptype))
	  (define-key gud-menu-map [run]     '("Run" . gud-run))
	  (define-key gud-menu-map [display]
	    '("Display expression".gud-display))
	  (define-key gud-menu-map [Output]
	    '("Debugger Output" .
	      (lambda() (interactive)
		(display-buffer glide-debug-buffer t))))
	  ))
    
    ;;  Modifies the menus
    (glide-debug-menus)

    ;;  Set up the menu in the debugger window
    (gud-make-debug-menu)

    ;;  Autoscroll the debugger window
    (set (make-local-variable 'comint-scroll-to-bottom-on-input) t)
    (set (make-local-variable 'comint-scroll-to-bottom-on-output) t)
  
    ;;  Contextual menus
    (set 'glide-debug-buffer (current-buffer))

    ;;  Display the debugger buffer when using print, since the output is
    ;;  always sent there
    (defun gud-print (arg) (interactive "p")
      (display-buffer glide-debug-buffer t)
      (gud-call "print %e" arg))
    
    ;;  Defines some new commands
    (defun gud-run (arg)
      "Run the application."
      (interactive "p")
      (gud-call "run" arg))
    (defun gud-break (arg)
      "Run the application."
      (interactive "p")
      (gud-call "break %f:*:%l" arg))
    (defun gud-display (arg)
      "Display the expression at point."
      (interactive "p")
      (gud-call "display %e" arg))
    (defun gud-ptype (arg)
      "print the type of the expression at point."
      (interactive "p")
      (gud-call "ptype %e" arg))
    
    ;;  Enhanced ddd support
    ;; If we are using ddd or gvd, modify the command to display data
    (let ((cmd (ada-xref-get-project-field 'debug_cmd)))
      (if (string-match "\\<gvd\\>" cmd)
	  (glide-debug-set-gvd-support)
	(if (string-match "\\<ddd\\>" cmd)
	    (glide-debug-set-ddd-support))))
   )
  
  ;;  Activate our own filter for the output (enhanced breakpoint support)

  (if (boundp 'running-xemacs)
      (progn
	(set 'glide-debug-default-filter (symbol-function gud-marker-filter))
	(funcall (symbol-function 'gud-overload-functions)
	 '((gud-marker-filter . glide-debug-marker-filter))))
    
    (set 'glide-debug-default-filter gud-marker-filter)
    (set 'gud-marker-filter 'glide-debug-marker-filter)
    ))


;;  set the hooks for the initialization
(add-hook 'gdb-mode-hook 'glide-debug-initialize)
(add-hook 'jdb-mode-hook 'glide-debug-initialize)
(set 'ada-contextual-menu
     (glide-debug-contextual-menu ada-contextual-menu))

(provide 'glide-debug)