------------------------------------------------------------------------------
--                                                                          --
--                         GNAT COMPILER COMPONENTS                         --
--                                                                          --
--                                G L I D E I N T                           --
--                                                                          --
--                                 B o d y                                  --
--                                                                          --
--                            $Revision: 1.16 $                              --
--                                                                          --
--          Copyright (C) 1992-2000 Free Software Foundation, Inc.          --
--                                                                          --
-- GNAT is free software;  you can  redistribute it  and/or modify it under --
-- terms of the  GNU General Public License as published  by the Free Soft- --
-- ware  Foundation;  either version 2,  or (at your option) any later ver- --
-- sion.  GNAT is distributed in the hope that it will be useful, but WITH- --
-- OUT ANY WARRANTY;  without even the  implied warranty of MERCHANTABILITY --
-- or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU General Public License --
-- for  more details.  You should have  received  a copy of the GNU General --
-- Public License  distributed with GNAT;  see file COPYING.  If not, write --
-- to  the Free Software Foundation,  59 Temple Place - Suite 330,  Boston, --
-- MA 02111-1307, USA.                                                      --
--                                                                          --
-- GNAT was originally developed  by the GNAT team at  New York University. --
-- It is now maintained by Ada Core Technologies Inc (http://www.gnat.com). --
--                                                                          --
------------------------------------------------------------------------------

--  This package is the startup program for Glide, the GNAT leveraged
--  Integrated Deveveloppment environment.
--  It starts Emacs, after parsing the command line for special parameters

with Ada.Text_IO;           use Ada.Text_IO;
with Ada.Command_Line;      use Ada.Command_Line;
with GNAT.Command_Line;     use GNAT.Command_Line;
with GNAT.OS_Lib;           use GNAT.OS_Lib;
with Ada.Strings.Unbounded; use Ada.Strings.Unbounded;

procedure GlideInt is

   Running_On_Windows : Boolean := False;

   Color_Highlight : Boolean := True;
   Speedbar_Mode   : Boolean := True;
   Vi_Mode         : Boolean := False;
   --  Default Values, can be changed depending on the system

   Program_Name    : Unbounded_String;
   --  Name of the editor (could be Emacs, XEmacs or runemacs)


   --  Before spawning an Emacs subprocess, we create an array with all
   --  the switches.
   --  Some switches need to be given on the command line before loading
   --  the file glide.el (for instance to modify the load-path,...), whereas
   --  some need to be passed after loading glide (because they reference
   --  functions defined in it).
   --  The array of switches is Emacs_Switches, and it is organized like this:
   --    Initial_Index .. Preload_Switches - 1
   --                 => Switches before loading glide
   --    Preload_Switches .. Preload_Switches + Default_Switches - 1
   --                 => Default switches for Emacs (load glide, ...)
   --    Preload_Switches .. Emacs_Number
   --                 => Switches after loading glide (file names, ...)
   --  The reason we create this array instead of creating a string is because
   --  some switches, like filenames, could include spaces

   Preload_Switches : constant Positive := 3;
   Default_Switches : constant Positive := 2;
   Initial_Index    : Positive := Preload_Switches;
   Emacs_Number     : Natural := Preload_Switches + Default_Switches - 1;
   Emacs_Switches   : Argument_List (1 .. Argument_Count + Emacs_Number);

   Update_Mode     : Boolean := False;

   Init_Func       : Unbounded_String;
   --  Lisp expression to evaluate when starting Emacs

   Local_Dir       : Unbounded_String;
   --  Directory to load the files from

   New_Extension   : Unbounded_String;
   --  List of new possible extensions for Ada files (separated by spaces)

   Install_Dir     : Unbounded_String;
   --  Directory in which the ada-mode files have been installed

   Emacs_Pid       : GNAT.OS_Lib.Process_Id;

   procedure File_Copy (Name : String; Directory : String);
   --  Copy the file Name to the directory.

   procedure Update;
   --  Update the ada-mode installation, when it has previously been installed
   --  in Site_Lisp

   procedure Usage;
   --  Print usage information for this program, and exit

   procedure Save_Switch (S : String);
   --  Save the switch and transfer it to emacs

   procedure Parse_Cmd_Line;
   --  Parse the command line

   function Protect (S : String) return String;
   --  Protect the special characters in the string S (the ones that can not
   --  be passed as is to an external shell)

   function Is_Windows return Boolean;
   --  Return True if we are running on windows

   ---------------
   -- File_Copy --
   ---------------

   procedure File_Copy (Name : String; Directory : String) is
   begin
      Put_Line ("File_Copy not implemented yet.");
      Put_Line ("Check in the Runtime library");
   end File_Copy;

   --------------------
   -- Parse_Cmd_Line --
   --------------------

   procedure Parse_Cmd_Line is
   begin
      loop
         case Getopt
           ("color emacs: nospeedbar ext: h help local: nocolor update "
            & "vi standard platform")
         is
            when Ascii.NUL => exit;

            when 'c' =>  --  only choice is "color"
               Color_Highlight := True;

            when 'e' =>
               if Full_Switch = "ext" then
                  Append (New_Extension, Parameter & " ");
               elsif Full_Switch = "emacs" then
                  Save_Switch (Parameter);
               end if;

            when 'h' =>  --  only choice is "help"
               Usage;

            when 'l' =>  --  only choice is "local"
               Local_Dir := To_Unbounded_String (Parameter);

            when 'n' =>
               if Full_Switch = "nocolor" then
                  Color_Highlight := False;
               elsif Full_Switch = "nospeedbar" then
                  Speedbar_Mode := False;
               end if;

            when 'p' | 's' =>  --  only choice is "platform"
               Put_Line
               ("Switches -p and -s are no longer supported in Glide");

            when 'u' =>  --  only choice is "update"
               Update_Mode := True;

            when 'v' =>  --  only choice is "vi"
               Vi_Mode := True;

            when others =>
               raise Program_Error;
         end case;
      end loop;

      --  First argument is the installation directory
      Install_Dir :=
        To_Unbounded_String (Get_Argument (Do_Expansion => False));
      declare
         S : String := To_String (Install_Dir);
      begin
         if S (S'Last) /= '/' then
            Append (Install_Dir, '/');
         end if;

         --  if the path includes a ':' as the second character, we are running
         --  on windows

         if S (S'First + 1) = ':' then
            Running_On_Windows := True;
         end if;
      end;


      --  Second argument is the executable name
      Program_Name :=
        To_Unbounded_String (Get_Argument (Do_Expansion => False));

      --  Others are the file names to edit
      loop
         declare
            S : constant String := Get_Argument (Do_Expansion => True);

         begin
            exit when S'Length = 0;
            Emacs_Number := Emacs_Number + 1;
            Emacs_Switches (Emacs_Number) := new String'(S);
         end;
      end loop;

   exception
      when GNAT.Command_Line.Invalid_Switch =>
         Put_Line ("Invalid switch : " & Parameter);
         Usage;

      when GNAT.Command_Line.Invalid_Parameter =>
         Put_Line ("Parameter missing for : " & Parameter);
         Usage;

   end Parse_Cmd_Line;

   -----------------
   -- Save_Switch --
   -----------------

   procedure Save_Switch (S : String) is
   begin
      Emacs_Number := Emacs_Number + 1;
      Emacs_Switches (Emacs_Number) := new String'(S);
   end Save_Switch;

   -------------
   -- Protect --
   -------------

   function Protect (S : String) return String is
      Value : String := S;
   begin
      for I in Value'Range loop
         if Value (I) = '\' then
            Value (I) := '/';
         end if;
      end loop;
      return Value;
   end Protect;

   ------------
   -- Update --
   ------------

   procedure Update is
      Ada_Mode_Files : constant array (Positive range <>)
        of GNAT.OS_Lib.String_Access
        := (new String'("ada-mode.el"),
            new String'("ada-xref.el"),
            new String'("ada-stmt.el"),
            new String'("ada-prj.el"),
            new String'("ada-support.el"),
            new String'("glide.el"),
            new String'("glide-debug.el"));
      To_Install : array (Ada_Mode_Files'Range) of Boolean
        := (others => False);
      Number_Of_Files : Natural := 0;
      Success : Boolean;

   begin

      --  Find the list of files to be installed

      for Index in Ada_Mode_Files'Range loop
         if Is_Regular_File (Ada_Mode_Files (Index).all) then
            To_Install (Index) := True;
            Number_Of_Files := Number_Of_Files + 1;
         end if;
      end loop;

      --  Check that it is possible to write the compiled file into the
      --  destination directory.
      --  Note that if the file does not exist but the directory is writable,
      --  this is ok.

      for Index in Ada_Mode_Files'Range loop
         if To_Install (Index) then
            declare
               F : String := To_String (Install_Dir)
                 & Ada_Mode_Files (Index).all & "c";
            begin
               if (Is_Regular_File (F) and not Is_Writable_File (F)) then
                  Put_Line ("You do not have permission to write "
                            & Ada_Mode_Files (Index).all
                            & "c in the directory");
                  Put_Line (To_String (Install_Dir));
                  Put_Line ("Please check the permissions");
                  OS_Exit (-1);
               end if;
            end;
         end if;
      end loop;

      --  Byte-Compile the files

      declare
         Switches : Argument_List (1 .. 8 + Number_Of_Files);

      begin
         Switches (1) := new String'("-q");
         Switches (2) := new String'("-batch");
         Switches (3) := new String'("-l");
         Switches (4) := new String'("ada-mode");
         Switches (5) := new String'("--eval");

         if Running_On_Windows then
            Switches (6) := new String'
              ("""(progn(setq byte-compile-warnings nil)"
               & "(add-to-list 'load-path \"".\""))""");
         else
            Switches (6) := new String'
              ("(progn(setq byte-compile-warnings nil)"
               & "(add-to-list 'load-path "".""))");
         end if;

         Switches (7) := new String'("-f");
         Switches (8) := new String'("batch-byte-compile");
         Number_Of_Files := 9;

         Ada.Text_IO.Put_Line ("Byte-compiling the files.");
         for Index in Ada_Mode_Files'Range loop
            if To_Install (Index) then
               Switches (Number_Of_Files)
                 := new String'(Ada_Mode_Files (Index).all);
               Number_Of_Files := Number_Of_Files + 1;
            end if;
         end loop;
         GNAT.OS_Lib.Spawn (Locate_Exec_On_Path (To_String (Program_Name)).all,
                            Switches,
                            Success);
      end;

      --  Copy and install the files
      --  We use Emacs to copy, since there is no standard and portable
      --  way of doing a copy directly from Ada or C.
      --  We also disable the backups that Emacs tries to do...

      declare
         Switches : Argument_List (1 .. 5);

      begin
         Switches (1) := new String'("-q");
         Switches (2) := new String'("-batch");

         Switches (4) := new String'("--eval");

         if Running_On_Windows then
            Switches (5) := new String'
              ("""(progn(setq backup-inhibited t)"
               & "(write-file \""" & To_String (Install_Dir) & "\""))""");
         else
            Switches (5) := new String'
              ("(progn(setq backup-inhibited t)"
               & "(write-file """ & To_String (Install_Dir) & """))");
         end if;

         Put_Line ("-- Installing the files");
         for Index in Ada_Mode_Files'Range loop
            if To_Install (Index) then
               Switches (3) := new String'(Ada_Mode_Files (Index).all & "c");
               GNAT.OS_Lib.Spawn
                 (Locate_Exec_On_Path (To_String (Program_Name)).all,
                  Switches,
                  Success);
            end if;
         end loop;
      end;

      OS_Exit (0);
   end Update;

   ----------------
   -- Is_Windows --
   ----------------

   function Is_Windows return Boolean is
   begin
      return Directory_Separator = '\';
   end Is_Windows;

   -----------
   -- Usage --
   -----------

   procedure Usage is
      use Ada.Text_IO;
   begin
      Put_Line ("GLIDE Copyright 1998-2002, Ada Core Technologies Inc.");
      Put_Line ("Usage: glide [switches] files");
      New_Line;
      Put_Line ("glide switches:");
      Put_Line (" -h              Print this help page");
      Put_Line (" -color          Activate the color highlighting");
      Put_Line (" -emacs switch   Pass the switch directly to Emacs");
      Put_Line (" -ext spec,body  Specify a new set of extension for "
                & "Ada files");
      Put_Line ("                 The default is:  -ext .ads,.adb");
      Put_Line (" -local dir      Load the ada-mode from dir");
      Put_Line (" -nocolor        Deactivate the color highlighting");
      Put_Line (" -nospeedbar     Do not launch the speedbar on startup");
      Put_Line (" -update         Update glide (from files in current"
                & " directory)");
      Put_Line (" -vi             Vi-emulation mode");
      New_Line;
      Put_Line (" -platform       Use platform-specific dialogs and key"
                & " bindings");
      Put_Line (" -standard       Use standard dialogs");
      OS_Exit (0);
   end Usage;

begin

   --  Parses the command line

   Parse_Cmd_Line;

   --  If asked by the user, update the installation of ada-mode, and exit
   if Update_Mode then
      Update;
   end if;

   --  Create the initialization function

   Init_Func := To_Unbounded_String
     ("(progn(load-library (symbol-name 'glide))");

   --  Add new Ada extensions if needed

   if Length (New_Extension) /= 0 then
      if Running_On_Windows then
         Append (Init_Func, "(glide-add-extensions \"""
                 & To_String (New_Extension) & "\"")");
      else
         Append (Init_Func, "(glide-add-extensions """
                 & To_String (New_Extension) & """)");
      end if;
   end if;

   if Color_Highlight then
      Append (Init_Func, "(glide-font-lock)");
   end if;

   if Speedbar_Mode then
      Append (Init_Func, "(glide-explorer)");
   end if;

   if Vi_Mode then
      Append (Init_Func, "(glide-vi-mode)");
   end if;

   Append (Init_Func, ")");

   --  Modify the load-path first

   if Length (Local_Dir) /= 0 then
      Emacs_Switches (Initial_Index - 2) := new String'("-L");
      Emacs_Switches (Initial_Index - 1) := new String'(To_String (Local_Dir));
      Initial_Index := Initial_Index - 2;
   end if;

   --  Now starts the editor
   Emacs_Switches (Preload_Switches + 0) := new String'("--eval");

   if Running_On_Windows then
      Emacs_Switches (Preload_Switches + 1)
        := new String'("""" & To_String (Init_Func) & """");
   else
      Emacs_Switches (Preload_Switches + 1)
        := new String'(To_String (Init_Func));
   end if;

   declare
      Prog : GNAT.OS_Lib.String_Access
        := Locate_Exec_On_Path (To_String (Program_Name));
      Success : Boolean;
   begin
      if Prog = null then
         Put_Line (To_String (Program_Name) & " not found in the path...");
      else

         --  Graphic mode ? Start Emacs in background

         if Is_Windows or else Getenv ("DISPLAY").all /= "" then
            Emacs_Pid := GNAT.OS_Lib.Non_Blocking_Spawn
              (Prog.all,
               Emacs_Switches (Initial_Index .. Emacs_Number));
         else
            declare
               New_Switches : Argument_List (Initial_Index .. Emacs_Number+ 1);
            begin
               New_Switches (Initial_Index .. Emacs_Number) :=
                 Emacs_Switches (Initial_Index .. Emacs_Number);
               New_Switches (Emacs_Number + 1) := new String'("-nw");
               GNAT.OS_Lib.Spawn (Prog.all, New_Switches, Success);
            end;
         end if;
      end if;
   end;
end GlideInt;


